/// This file is a part of media_kit (https://github.com/alexmercerind/media_kit).
///
/// Copyright © 2021 & onwards, Hitesh Kumar Saini <saini123hitesh@gmail.com>.
/// All rights reserved.
/// Use of this source code is governed by MIT license that can be found in the LICENSE file.
import 'dart:ffi';
import 'dart:collection';
import 'package:ffi/ffi.dart';

import 'package:media_kit/src/isolates.dart';

/// {@template android_asset_loader}
///
/// AndroidAssetLoader
/// ------------------
///
/// This class is used to assets bundled with the application on Android.
/// The implementation depends on the mediakitandroidhelper library.
///
/// Learn more: https://github.com/media-kit/media-kit-android-helper
///
/// {@endtemplate}
abstract class AndroidAssetLoader {
  /// Copies an asset bundled with the application to the external files directory & returns it absolute path.
  static Future<String> load(String asset) async {
    if (_loaded.containsKey(asset)) {
      return _loaded[asset]!;
    }

    // Run on another [Isolate] to avoid blocking the UI.
    final path = await compute(loadSync, asset);

    _loaded[asset] = path;

    return path;
  }

  /// Copies an asset bundled with the application to the external files directory & returns it absolute path.
  @pragma('vm:entry-point')
  static String loadSync(String asset) {
    if (_loaded.containsKey(asset)) {
      return _loaded[asset]!;
    }

    final lib = DynamicLibrary.open('libmediakitandroidhelper.so');
    final fn = lib.lookupFunction<FnCXX, FnDart>(
      'MediaKitAndroidHelperCopyAssetToExternalFilesDir',
    );

    final name = asset.toNativeUtf8();
    final result = List.generate(4096, (index) => ' ').join('').toNativeUtf8();

    fn.call(
      name.cast(),
      result.cast(),
    );

    final path = result.cast<Utf8>().toDartString().trim();

    calloc.free(name);
    calloc.free(result);

    _loaded[asset] = path;

    return path;
  }

  /// Stores the names of previously loaded assets. This avoids redundant FFI calls.
  static final HashMap<String, String> _loaded = HashMap<String, String>();
}

// Type definitions for native functions in the shared library.

// C/C++:

typedef FnCXX = Void Function(Pointer<Utf8> asset, Pointer<Utf8> result);

// Dart:

typedef FnDart = void Function(Pointer<Utf8> asset, Pointer<Utf8> result);
